
#include <iostream>
#include <string>
#include <iomanip>
#include <CImg.h>
#include <Timer.hpp>

// Sequential kernel
void kernel_h(const unsigned int width, const unsigned int height, const unsigned int spectrum, const unsigned char * inputImage, unsigned char * outputImage);
// OpenCL kernel
void kernel_d(const unsigned int width, const unsigned int height, const unsigned int spectrum, const unsigned char * inputImage, unsigned char * outputImage);


int main(int argc, char * argv[]) {
  if ( argc != 2 ) {
    std::cerr << argv[0] << " <input_image>" << std::endl;
    return 1;
  }

  // Load input image
  cimg_library::CImg< unsigned char > inputImage = cimg_library::CImg< unsigned char >(argv[1]);
  // Prepare output
  cimg_library::CImg< unsigned char > outputImage_h = cimg_library::CImg< unsigned char >(inputImage.height(), inputImage.width(), 1, inputImage.spectrum());
  cimg_library::CImg< unsigned char > outputImage_d = cimg_library::CImg< unsigned char >(inputImage.height(), inputImage.width(), 1, inputImage.spectrum());

  // Lauch kernels
  kernel_h(inputImage.width(), inputImage.height(), inputImage.spectrum(), inputImage.data(), outputImage_h.data());
  kernel_d(inputImage.width(), inputImage.height(), inputImage.spectrum(), inputImage.data(), outputImage_d.data());
  
  // Correctness check
  long long unsigned int wrongItems = 0;

  for ( int x = 0; x < outputImage_h.width(); x++ ) {
    for ( int y = 0; y < outputImage_h.height(); y++ ) {
      for ( int c = 0; c < outputImage_h.spectrum(); c++ ) {
        if ( outputImage_h(x, y, 0, c) != outputImage_d(x, y, 0, c) ) {
          wrongItems++;
        }
      }
    }
  }
  if ( wrongItems > 0 ) {
    std::cout << "Wrong: \t\t" << wrongItems << std::fixed << std::setprecision(2) << " (" << (wrongItems * 100.0) / (inputImage.width() * inputImage.height() * inputImage.spectrum()) << "%)" << std::endl;
  }

  return 0;
}

void kernel_h(const unsigned int width, const unsigned int height, const unsigned int spectrum, const unsigned char * inputImage, unsigned char * outputImage) {
  LOFAR::NSTimer kernelTimer("Kernel", false, false);

  kernelTimer.start();
  // Kernel
  for ( unsigned int c = 0; c < spectrum; c++ ) {
    for ( unsigned int y = 0; y < height; y++ ) {
      for ( unsigned int x = 0; x < width; x++ ) {
        outputImage[(c * width * height) + (x * height) + y] = inputImage[(c * width * height) + (y * width) + x];
      }
    }
  }
  // /Kernel
  kernelTimer.stop();

  // Print performance metrics
  std::cout << "Sequential" << std::endl;
  std::cout << "Time: \t\t" << std::fixed << std::setprecision(6) << kernelTimer.getElapsed() << std::endl;
  std::cout << "GFLOP/s: \t" << std::fixed << std::setprecision(3) << "-" << std::endl;
  std::cout << "GB/s: \t\t" << std::fixed << std::setprecision(3) << (2 * sizeof(unsigned char) * static_cast< long long unsigned int >(width) * height * spectrum) / 1000000000.0 / kernelTimer.getElapsed() << std::endl;
}

void kernel_d(const unsigned int width, const unsigned int height, const unsigned int spectrum, const unsigned char * inputImage, unsigned char * outputImage) {
  LOFAR::NSTimer kernelTimer("Kernel", false, false);
  LOFAR::NSTimer memoryTimer("Memory", false, false);
  LOFAR::NSTimer globalTimer("Global", false, false);

  kernelTimer.start();
  // Kernel
  for ( unsigned int c = 0; c < spectrum; c++ ) {
    for ( unsigned int y = 0; y < height; y++ ) {
      for ( unsigned int x = 0; x < width; x++ ) {
        outputImage[(c * width * height) + (x * height) + y] = inputImage[(c * width * height) + (y * width) + x];
      }
    }
  }
  // /Kernel
  kernelTimer.stop();

  // Print performance metrics
  std::cout << "OpenCL" << std::endl;
  std::cout << "Time (g): \t" << std::fixed << std::setprecision(6) << globalTimer.getElapsed() << std::endl;
  std::cout << "Time (m): \t" << std::fixed << std::setprecision(6) << memoryTimer.getElapsed() << std::endl;
  std::cout << "Time (k): \t" << std::fixed << std::setprecision(6) << kernelTimer.getElapsed() << std::endl;
  std::cout << "GFLOP/s: \t" << std::fixed << std::setprecision(3) << "-" << std::endl;
  std::cout << "GB/s: \t\t" << std::fixed << std::setprecision(3) << "-" << std::endl;
}

